package miscc;

import java.util.*;
import java.io.PrintWriter;

/**
 * Control-flow graph (CFG). A CFG is composed of nodes, which are
 * basic blocks containing non-jumping instructions, and edges, which
 * are labelled with jump instructions.
 */

public class CFG {
    public final String name;

    private LinkedList/*<Node>*/ nodes = new LinkedList();
    private ArrayList/*<Node>*/ labels = new ArrayList();
    private HashMap/*<Node, List<Edge>>*/ edges = new HashMap();
    private HashMap/*<Node, List<Edge>>*/ revEdges = new HashMap();

    public CFG(String name) {
	this.name = name;
    }

    public class Node {
        /** Instructions composing this node (basic block) */
        public Code code = new Code();
        /** Label number associated with this node */
	public final int label;

	public Node() {
	    label = labels.size();
            labels.add(this);
	}

        public void emit(Instruction i) {
            code.emit(i);
        }

        public void emit(int opcode) {
            code.emit(opcode);
        }

        public void emit(int opcode, int c) {
            code.emit(opcode, c);
        }

        public void emit(int opcode, String c) {
            code.emit(opcode, c);
        }

        public void emit(int opcode, int a, int c) {
            code.emit(opcode, a, c);
        }

        public void emit(int opcode, int a, String c) {
            code.emit(opcode, a, c);
        }

        public void emit(int opcode, int a, int b, int c) {
            code.emit(opcode, a, b, c);
        }

    }

    public class Edge {
        public final Node from, to;
	public final Instruction branch;
        public Edge(Node from, Node to, int opcode, int reg) {
            this.from = from;
            this.to = to;
	    this.branch = Instruction.make(opcode,
                                                      reg,
                                                      name + "_" + to.label);
	}
    }

    public Node addNewNode() {
        Node newNode = new Node();
        nodes.add(newNode);
        return newNode;
    }

    public Node getNode(int label) {
	return (Node)labels.get(label);
    }

    public List getNodes() {
        return Collections.unmodifiableList(nodes);
    }

    public Edge addNewEdge(Node from, Node to, int opcode, int reg) {
        Edge newEdge = new Edge(from, to, opcode, reg);

        LinkedList fromL = (LinkedList)edges.get(from);
        if (fromL == null) {
            fromL = new LinkedList();
            edges.put(from, fromL);
        }
        fromL.add(newEdge);

        LinkedList toL = (LinkedList)revEdges.get(to);
        if (toL == null) {
            toL = new LinkedList();
            revEdges.put(to, toL);
        }
        toL.add(newEdge);

        return newEdge;
    }

    public Instruction getLastInstruction(Node node) {
        List edgesFrom = getEdgesFrom(node);
        if (edgesFrom.isEmpty()) {
            List instrs = node.code.instructions();
            return (Instruction)instrs.get(instrs.size() - 1);
        } else
	    return ((Edge)edgesFrom.get(0)).branch;
    }


    public List getEdgesFrom(Node node) {
        List edgesFrom = (List)edges.get(node);
        if (edgesFrom != null)
            return Collections.unmodifiableList((List)edgesFrom);
        else
            return Collections.EMPTY_LIST;
    }

    public List/*<Node>*/ successors(Node node) {
        List/*<Node>*/ succs = new LinkedList();
        Iterator edgesIt = getEdgesFrom(node).iterator();
        while (edgesIt.hasNext()) succs.add(((Edge)edgesIt.next()).to);
        return succs;
    }

    public List getEdgesTo(Node node) {
	List edgesTo = (List)revEdges.get(node);
        if (edgesTo != null)
            return Collections.unmodifiableList((List)edgesTo);
        else
            return Collections.EMPTY_LIST;
    }

    public List/*<Node>*/ predecessors(Node node) {
        List/*<Node>*/ preds = new LinkedList();
        Iterator edgesIt = getEdgesTo(node).iterator();
        while (edgesIt.hasNext()) preds.add(((Edge)edgesIt.next()).from);
        return preds;
    }

    public static interface InstrVisitor {
	public void visit(Instruction i);
    }

    public void visitNodeForward(Node node, InstrVisitor visitor) {
        Iterator instrIt = node.code.instructions().iterator();
        while (instrIt.hasNext())
	    visitor.visit((Instruction)instrIt.next());

	Iterator edgeIt = getEdgesFrom(node).iterator();
	while (edgeIt.hasNext()) {
	    Edge edge = (Edge)edgeIt.next();
	    visitor.visit(edge.branch);
	}
    }

    public void visitNodeBackward(Node node, InstrVisitor visitor) {
	Iterator edgeIt = getEdgesFrom(node).iterator();
	while (edgeIt.hasNext()) {
	    Edge edge = (Edge)edgeIt.next();
	    visitor.visit(edge.branch);
	}

        List instrs = node.code.instructions();
        ListIterator instrIt = instrs.listIterator(instrs.size());
        while (instrIt.hasPrevious())
	    visitor.visit((Instruction)instrIt.previous());
    }

    /** Visit the CFG in depth-first search. */
    public void visitDFS(Visitor visitor) {
	if (!nodes.isEmpty())
	    visitDFS(visitor, (Node)nodes.getFirst(), new HashSet());
    }

    private void visitDFS(Visitor visitor, Node root, HashSet visited) {
        if (!visited.contains(root)) {
            visited.add(root);
            visitor.pre(root);

            Iterator outIt = getEdgesFrom(root).iterator();
            while (outIt.hasNext()) {
                Edge outEdge = (Edge)outIt.next();
                visitDFS(visitor, outEdge.to, visited);
            }
            visitor.post(root);
        }
    }

    /** Visit the CFG in reverse pre-order. */
    public void visitRPO(Visitor visitor) {
        final LinkedList/*<Node>*/ nodes = new LinkedList();
        visitDFS(new CFG.EmptyVisitor() {
                public void post(CFG.Node node) { nodes.addFirst(node); }
            });
	Iterator nodesIt = nodes.iterator();
	while (nodesIt.hasNext()) {
            Node n = (Node)nodesIt.next();
            visitor.pre(n);
            visitor.post(n);
	}
    }

    /** Visitor for DFS traversal. */
    public static interface Visitor {
        /** Method called before visiting the children of the node */
        public void pre(Node node);
        /** Method called after visiting the children of the node */
        public void post(Node node);
    }

    /** Default visitor, with empty pre and post methods. */
    public static class EmptyVisitor implements Visitor {
        public void pre(Node node) {}
        public void post(Node node) {}
    }
}
