package miscc;

import java.util.*;

abstract public class ControlFlowAnalysis {
    protected final CFG cfg;

    private final HashMap/*<Node, Object>*/ in = new HashMap();
    private final HashMap/*<Node, Object>*/ out = new HashMap();

    public ControlFlowAnalysis(CFG cfg) {
        this.cfg = cfg;
    }

    public Object in(CFG.Node n) {
        return in.get(n);
    }

    public Object out(CFG.Node n) {
        return out.get(n);
    }

    public void analyseBackwards() {
        final HashMap/*<Node, Transfer>*/ transferFunctions =
            new HashMap();

        /* Initialise the analysis.
           Calculate transfer functions for each BB.
           Store the result in transferFunctions,
           a map Node->Transfer. */
        cfg.visitDFS(new CFG.EmptyVisitor() {
                public void pre(CFG.Node node) {
                    /* Calculate transfer functions for this node. */
                    transferFunctions.put(node, calculateTransfer(node));
                    /* Initialise in and out.*/
                    in.put(node, newValue());
                    out.put(node, newValue());
                }
            });

        /* Fix-point iteration **/
        LinkedList/*<Node>*/ worklist = new LinkedList(cfg.getNodes());
        while (! worklist.isEmpty()) {
            /* Take an element (n) from the worklist. */
            CFG.Node n = (CFG.Node) worklist.removeFirst();
            Object oldIn = cloneValue(in(n));
            Object newOut = out(n);

            /* Apply convergence function: U(succ(n).in) */
            Iterator succIt = cfg.successors(n).iterator();
            while (succIt.hasNext())
                newOut = mergeValues(newOut, in((CFG.Node)succIt.next()));
            out.put(n, newOut);

            /* Apply transfer function */
            Object newIn =
                (((Transfer)transferFunctions.get(n)).apply(newOut));
            in.put(n, newIn);

            /* If "in" changed, add predecessors to worklist. */
            if (! newIn.equals(oldIn)) {
                List preds = cfg.predecessors(n);
                preds.removeAll(worklist);
                worklist.addAll(preds);
            }
        }
    }

    /** Create a new analysis value */
    abstract protected Object newValue();
    /** Clone an existing analysis value */
    abstract protected Object cloneValue(Object v);
    /** Merge two analysis values */
    abstract protected Object mergeValues(Object v1, Object v2);

    abstract protected Transfer calculateTransfer(CFG.Node node);

    protected abstract class Transfer {
        abstract Object apply(Object input);
        public String toString() { return "Object -> Object"; }
    }
}
