package miscc;

import java.util.*;
import miscc.Instruction;

class Frame {
    public void rewriteCalls(final CFG cfg) {
        final Liveness liveness = new Liveness(cfg);
        cfg.visitDFS(new CFG.EmptyVisitor() {
                public void pre(CFG.Node node) {
                    Code code = node.code;
                    Code newCode = new Code();
                    BitSet liveOut = liveness.liveOut(node);

                    Iterator edgeIt = cfg.getEdgesFrom(node).iterator();
                    while (edgeIt.hasNext()) {
                        Instruction instr = ((CFG.Edge)edgeIt.next()).branch;
                        liveOut.andNot(instr.defs());
                        liveOut.or(instr.uses());
                    }

                    List instrs = code.instructions();
                    ListIterator instrIt =
                        instrs.listIterator(code.instructions().size());
                    while (instrIt.hasPrevious()) {
                        Instruction instr = (Instruction)instrIt.previous();

                        liveOut.andNot(instr.defs());
                        if (instr.opcode == PseudoInstruction.PCALL) {
                            PCALL pInstr = (PCALL)instr;
                            rewriteCall(instr,
                                        new CALL(pInstr.addrReg),
                                        pInstr.resReg,
                                        liveOut,
                                        newCode);
                        } else if (instr.opcode == PseudoInstruction.PJSR) {
                            PJSR pInstr = (PJSR)instr;
                            rewriteCall(instr,
                                        new JSR(pInstr.label),
                                        pInstr.resReg,
                                        liveOut,
                                        newCode);
                        } else
                            newCode.addFirst(instr);
                        liveOut.or(instr.uses());
                    }
                    node.code = newCode;
                }
            });
    }

    private void rewriteCall(Instruction instr,
                             Instruction newInstr,
                             int resReg,
                             BitSet liveOut,
                             Code newCode) {
        LinkedList regsToSave;
        List args;
        Iterator argIt;
        Iterator regIt;

        // We are traversing the code backwards,
        // so we must generate the code backwards.

        // Restore live regs
        regsToSave = new LinkedList();
        for(int i = liveOut.nextSetBit(1);
            i >= 0 && i < 30;
            i = liveOut.nextSetBit(i+1)) {
            newCode.addFirst(new POP(i, RISC.SP, RISC.WORD_SIZE));
            regsToSave.addFirst(new Integer(i));
        }

        // Move result to target reg.
        newCode.addFirst(new ADDI(resReg, RISC.RES, 0));

        // Call function
        newCode.addFirst(newInstr);

        // Push arguments.
        args = ((PseudoCall) instr).args;
        pushArgs(args.iterator(), newCode);

        // Save live regs.
        regIt = regsToSave.iterator();
        while (regIt.hasNext()) {
            int reg = ((Integer)regIt.next()).intValue();
            newCode.addFirst(new PSH(reg, RISC.SP, RISC.WORD_SIZE));
        }
    }

    private void pushArgs(Iterator argIt, Code newCode) {
        if (argIt.hasNext()) {
            int arg = ((Integer)argIt.next()).intValue();
            pushArgs(argIt, newCode);
            newCode.addFirst(new PSH(arg, RISC.SP, RISC.WORD_SIZE));
        }
    }
}

