/*  miscc -- a compiler for misc
 *    Code generator
 *
 *    Michel Schinz
 *    $Id: Generator.java,v 1.11 2005/01/17 09:54:32 schinz Exp $
 */

package miscc;

import miscc.Tree.*;
import miscc.Instruction;

import java.util.List;
import java.util.LinkedList;
import java.util.BitSet;

public class Generator {
    private final Gen genVisitor;
    private final GenLoad genLoadVisitor;
    private final GenCond genCondVisitor;

    private final RegisterPool regPool;

    private final List/*<CFG>*/ cfgs;

    // Context
    private CFG currCFG;
    private CFG.Node currNode;

    private int frameSize = 0;

    public Generator(RegisterPool regPool) {
        this.regPool = regPool;
        this.genVisitor = new Gen();
        this.genLoadVisitor = new GenLoad();
        this.genCondVisitor = new GenCond();

        this.cfgs = new LinkedList();
    }

    public Generator() {
        this(new RegisterPool(32, Integer.MAX_VALUE));
    }

    public CFG[] generate(Tree.Program prog) {
        Tree[] defs = prog.defs;
        for (int i = 0; i < defs.length; ++i)
            genVisitor.caseFunDef((FunDef)defs[i]);
        currCFG = addNewCFG("_Main");
        currNode = currCFG.addNewNode();
        gen(prog.body);
        currNode.emit(RISC.RET, 0);
        return (CFG[])cfgs.toArray(new CFG[cfgs.size()]);
    }

    private CFG addNewCFG(String name) {
        CFG newCFG = new CFG(name);
        cfgs.add(newCFG);
        return newCFG;
    }

    private void gen(Tree tree) {
        genVisitor.generate(tree);
    }

    private void gen(Tree[] trees) {
        for (int i = 0; i < trees.length; ++i)
            gen(trees[i]);
    }

    private void genLoad(Tree tree, int targetReg) {
        assert !regPool.isFree(targetReg)
            : "invariant violated: target register R" + targetReg
            + " not allocated";
        genLoadVisitor.generate(tree, targetReg);
    }

    private void genCond(Tree tree, CFG.Node contTrue, CFG.Node contFalse) {
        genCondVisitor.generate(tree, contTrue, contFalse);
        currNode = null;        // genCond always terminates current block
    }

    /**
     * Visitor generating code for statements, in a general sense,
     * that is all constructs which do not have a value.
     */
    private class Gen extends DefaultVisitor implements Visitor, RISC {
        public void generate(Tree tree) {
            tree.apply(this);
        }

        public void caseProgram(Program tree) {
            throw new Error("no code to generate for " + tree);
        }

        public void caseFunDef(FunDef tree) {
            currCFG = addNewCFG(((Symbol.Fun)tree.sym).label);
            currNode = currCFG.addNewNode();

            // Prologue
            gen(tree.args);
            frameSize += WORD_SIZE;
            currNode.emit(PSH,LNK, SP, WORD_SIZE);

            // Body
            int resReg = regPool.get();
            genLoad(tree.body, resReg);
            currNode.emit(ADDI, RES, resReg, 0);
            regPool.free(resReg);

            // Epilogue
            int popSize = WORD_SIZE * (1 + tree.args.length);
            currNode.emit(POP, LNK, SP, popSize);
            frameSize -= popSize;
            assert frameSize == 0
                : "non-empty frame (size " + frameSize + ")";
            currNode.emit(RET, LNK);
        }

        public void caseFormal(Formal tree) {
            frameSize += WORD_SIZE;
            tree.sym.offset = frameSize;
        }

        public void caseBasicType(BasicType tree) {
            throw new Error("no code to generate for " + tree);
        }

        public void caseListType(ListType tree) {
            throw new Error("no code to generate for " + tree);
        }

        public void caseFunType(FunType tree) {
            throw new Error("no code to generate for " + tree);
        }

        public void caseVarDef(VarDef tree) {
            int tmpReg = regPool.get();
            genLoad(tree.init, tmpReg);
            frameSize += WORD_SIZE;
            currNode.emit(PSH, tmpReg, SP, WORD_SIZE);
            tree.var.sym.offset = frameSize;
        }

        public void caseWhile(While tree) {
            CFG.Node condNode = currCFG.addNewNode();
            currCFG.addNewEdge(currNode, condNode, BEQ, 0);
            CFG.Node afterNode = currCFG.addNewNode();
            CFG.Node bodyNode = currCFG.addNewNode();
            currNode = condNode;
            genCond(tree.cond, bodyNode, afterNode);
            currNode = bodyNode;
            gen(tree.body);
            currCFG.addNewEdge(currNode, condNode, BEQ, 0);
            currNode = afterNode;
        }

        public void caseExec(Exec tree) {
            gen(tree.expr);
        }

        public void caseDefault(Tree tree) {
            int tmpReg = regPool.get();
            genLoad(tree, tmpReg);
        }
    }

    /**
     * Visitor generating code for expressions, that is constructs
     * which produce a value.
     */
    private class GenLoad extends DefaultVisitor implements Visitor, RISC {
        private int targetReg;

        public void generate(Tree tree, int targetReg) {
            int bkpTargetReg = this.targetReg;
            this.targetReg = targetReg;
            tree.apply(this);
            this.targetReg = bkpTargetReg;
        }

        public void caseBody(Body tree) {
            int beginFrameSize = frameSize;
            gen(tree.stats);
            genLoad(tree.expr, targetReg);
            int endFrameSize = frameSize;
            if (endFrameSize > beginFrameSize) {
                currNode.emit(ADDI, SP, SP, endFrameSize - beginFrameSize);
                frameSize -= endFrameSize - beginFrameSize;
            }
        }

        public void caseIf(If tree) {
            CFG.Node thenNode = currCFG.addNewNode();
            CFG.Node elseNode = currCFG.addNewNode();
            CFG.Node afterNode = currCFG.addNewNode();
            genCond(tree.cond, thenNode, elseNode);
            currNode = thenNode;
            genLoad(tree.thenp, targetReg);
            currCFG.addNewEdge(currNode, afterNode, BEQ, 0);
            currNode = elseNode;
            genLoad(tree.elsep, targetReg);
            currCFG.addNewEdge(currNode, afterNode, BEQ, 0);
            currNode = afterNode;
        }

        public void caseAssign(Assign tree) {
            genLoad(tree.expr, targetReg);
            int realOffset = frameSize - tree.sym.offset;
            currNode.emit(STW, targetReg, SP, realOffset);
        }

        public void caseOperation(Operation tree) {
            switch (tree.op) {
            case Tree.ADD: case Tree.SUB:
            case Tree.MUL: case Tree.DIV: case Tree.MOD: {
                int leftReg = regPool.get();
                genLoad(tree.left, leftReg);
                int rightReg = regPool.get();
                genLoad(tree.right, rightReg);
                currNode.emit(opcode(tree.op), targetReg, leftReg, rightReg);
            } break;

            case Tree.CONS: {
                int tmpReg = regPool.get();
                genLoad(tree.left, tmpReg);
                currNode.emit(ADDI, targetReg, 0, 2 * WORD_SIZE);
                currNode.emit(SYSCALL, targetReg, targetReg, SYS_GC_ALLOC);
                currNode.emit(STW, tmpReg, targetReg, 0);
                genLoad(tree.right, tmpReg);
                currNode.emit(STW, tmpReg, targetReg, WORD_SIZE);
            } break;

            default:
                caseDefault(tree);
            }
        }

        public void caseListop(Listop tree) {
            switch (tree.op) {
            case Tree.HEAD:
            case Tree.TAIL: {
                int offset = WORD_SIZE * (tree.op == Tree.HEAD ? 0 : 1);
                genLoad(tree.args[0], targetReg);
                CFG.Node okNode = currCFG.addNewNode();
                CFG.Node errorNode = currCFG.addNewNode();
                currCFG.addNewEdge(currNode, okNode, BNE, targetReg);
                currCFG.addNewEdge(currNode, errorNode, BEQ, targetReg);
                errorNode.emit(ADDI, 1, ZERO, 1 + offset);
                errorNode.emit(SYSCALL, 1, ZERO, SYS_EXIT);
                currNode = okNode;
                currNode.emit(LDW, targetReg, targetReg, offset);
            } break;

            case Tree.ISEMPTY:
                caseDefault(tree);
                break;

            default:
                throw new Error("unknown list operator " + tree.op);
            }
        }

        public void caseIdent(Ident tree) {
            if (tree.sym instanceof Symbol.Var) {
                int realOffset = frameSize - ((Symbol.Var)tree.sym).offset;
                currNode.emit(LDW, targetReg, SP, realOffset);
            } else
                currNode.emit(LDA, targetReg, ((Symbol.Fun)tree.sym).label);
        }

        public void caseUnitLit(UnitLit tree) {
            // nothing to do
        }

        public void caseNilLit(NilLit tree) {
            currNode.emit(ADDI, targetReg, 0, 0);
        }

        public void caseIntLit(IntLit tree) {
            emitLoadConstant(targetReg, tree.value);
        }

        public void caseApply(Apply tree) {
            String funLabel = null;
            int funReg = -1;

            if ((tree.fun instanceof Ident)
                && ((Ident)tree.fun).sym instanceof Symbol.Fun)
                funLabel = ((Symbol.Fun)((Ident)tree.fun).sym).label;

            if (funLabel == null) {
                funReg = regPool.get();
                genLoad(tree.fun, funReg);
            }

            // evaluate arguments
	    LinkedList/*<Integer>*/ args = new LinkedList();
            for (int i = 0; i < tree.args.length; ++i) {
                int tmpReg = regPool.get();
                genLoad(tree.args[i], tmpReg);
		args.addLast(new Integer(tmpReg));
            }

            if (funLabel != null)
                currNode.emit(new PJSR(targetReg, args, funLabel));
            else
		currNode.emit(new PCALL(targetReg, args, funReg));
        }

        public void caseDefault(Tree tree) {
            genLoad(new If(tree.pos,
                           tree,
                           new IntLit(tree.pos, 1),
                           new IntLit(tree.pos, 0)),
                    targetReg);
        }

        private void emitLoadConstant(int reg, int value) {
            int highBits = value >> 16;
            int lowBits = value & 0xFFFF;
            if (highBits != 0) {
                currNode.emit(ORIU, reg, 0, highBits);
                currNode.emit(LSHI, reg, reg, 16);
                currNode.emit(ORIU, reg, reg, lowBits);
            } else
                currNode.emit(ORIU, reg, 0, lowBits);
        }

        private int opcode(int op) {
            switch (op) {
            case Tree.ADD: return ADD;
            case Tree.SUB: return SUB;
            case Tree.MUL: return MUL;
            case Tree.DIV: return DIV;
            case Tree.MOD: return MOD;
            default: throw new Error("unknown opcode: " + op);
            }
        }
    }

    /**
     * Visitor generating code for conditions, that is constructs
     * which jump somewhere on the result of a test.
     */
    private class GenCond extends DefaultVisitor implements Visitor, RISC {
        private CFG.Node trueCont, falseCont;

        public void generate(Tree tree, CFG.Node trueCont, CFG.Node falseCont){
            CFG.Node bkpTrueCont = this.trueCont;
            this.trueCont = trueCont;
            CFG.Node bkpFalseCont = this.falseCont;
            this.falseCont = falseCont;
            tree.apply(this);
            this.falseCont = bkpFalseCont;
            this.trueCont = bkpTrueCont;
        }

        public void caseOperation(Operation tree) {
            switch (tree.op) {
            case Tree.EQ: case Tree.NE: case Tree.LT:
            case Tree.LE: case Tree.GT: case Tree.GE:
                int leftReg = regPool.get();
                genLoad(tree.left, leftReg);
                int rightReg = regPool.get();
                genLoad(tree.right, rightReg);
                int targetReg = regPool.get();
                currNode.emit(CMP, targetReg, leftReg, rightReg);
                currCFG.addNewEdge(currNode,
                                   trueCont,
                                   opcode(tree.op, true),
                                   targetReg);
                currCFG.addNewEdge(currNode,
                                   falseCont,
                                   opcode(tree.op, false),
                                   targetReg);
                break;

            case Tree.AND:
                CFG.Node rightNode = currCFG.addNewNode();
                genCond(tree.left, rightNode, falseCont);
                currNode = rightNode;
                genCond(tree.right, trueCont, falseCont);
                break;

            case Tree.NOT:
                genCond(tree.left, falseCont, trueCont);
                break;

            default:
                caseDefault(tree);
            }
        }

        public void caseListop(Listop tree) {
            if (tree.op == Tree.ISEMPTY)
                genCond(tree.args[0], falseCont, trueCont);
            else
                caseDefault(tree);
        }

        public void caseDefault(Tree tree) {
            int tmpReg = regPool.get();
            genLoad(tree, tmpReg);
            currCFG.addNewEdge(currNode, trueCont, BNE, tmpReg);
            currCFG.addNewEdge(currNode, falseCont, BEQ, tmpReg);
        }

        private int opcode(int op, boolean negated) {
            switch (op) {
            case Tree.EQ : return (negated ? BEQ : BNE);
            case Tree.NE : return (negated ? BNE : BEQ);
            case Tree.LT : return (negated ? BLT : BGE);
            case Tree.LE : return (negated ? BLE : BGT);
            case Tree.GT : return (negated ? BGT : BLE);
            case Tree.GE : return (negated ? BGE : BLT);
            default      : throw new Error("invalid operator: " + op);
            }
        }
    }
}
