/*  Notes    :  * This is an implementation of
%%                "Linear Scan Register Allocation" by
%%                Massimiliano Poletto &amp; Vivek Sarkar described in
%%                ACM TOPLAS Vol 21, No 5, September 1999.
%%
*/

package miscc;

import java.util.*;

class LinearScan {
    private Liveness liveness;
    private CFG cfg;

    // We do not want to allocate temps to the following regs:
    //  0: Always 0
    //  1: Return value (Needs handling of pre-allocatated regs.)
    // 28: Temp reg - used for spilling. (Could use iteration instead)
    // 29: Temp reg - used for spilling. (Could use iteration instead)
    // 30: SP
    // 31: Link reg - Return address.
    private final RegisterPool regPool = new RegisterPool(2, 27);

    // intervals: Live intervals for temporaries
    // We start with an arbitrary sized array, it will grow as needed.
    private Interval[] intervals = new Interval[16];
    private int[] finalAllocations = null;

    public void scan(CFG cfg) {
	this.cfg=cfg;

	// -----------------------------------------------------
	// Step 1: Calculate liveness.
	liveness = new Liveness(cfg);

	// -----------------------------------------------------
	// Step 2: Calculate live-intervals for each register.
	cfg.visitRPO(new LiveIntervals());

	// -----------------------------------------------------
	// Step 3: Sort on start.
        Arrays.sort(intervals,
                    new Comparator() {
                        public int compare(Object o1, Object o2) {
                            if (o1 == o2) return 0;
                            if (o1 == null) return 1;
                            if (o2 == null) return -1;

                            Interval i1 = (Interval)o1;
                            Interval i2 = (Interval)o2;
                            if (i1.start == i2.start) {
                                if (i1.end == i2.end) {
                                    if (i1.reg == i2.reg) return 0;
                                    else if (i1.reg < i2.reg) return -1;
                                    else return 1;
                                } else if (i1.end < i2.end) return -1;
                                else return 1;
                            } else if (i1.start < i2.start) return -1;
                            else return 1;
                        }
                    });
	finalAllocations = new int[intervals.length];

	// -----------------------------------------------------
	// Step 4: Go through intervals and allocate regs.
        Tentative active = null;
        for (int i = 0; i < intervals.length && intervals[i] != null; ++i) {
	    Interval current = intervals[i];
	    if (active != null) active = active.expire(current.start);

	    int temp = current.reg;
	    // TODO: Add handling of precolored regs here if needed.
	    if (temp < 32) {
                finalAllocations[temp] = temp;
            } else {
		if (regPool.freeCount() > 0) { // Free reg available.
		    int reg = regPool.get();
		    // System.out.println("Allocaded  " + temp + "->" +reg);
		    Allocation a = new Allocation(temp, reg);
		    if (active == null) active = new Tentative(a, current.end);
		    else active = active.add(a, current.end);
		} else { // No free reg, spill needed.
		    active = active.spill(temp, current.end);
		}
	    }
	}
	// Move any unexpired tentative intervals to the finalAlloc.
	while (active != null) {active = active.expire(active.end+1);}

	// -----------------------------------------------------
	// Step 5: Rewrite the code.
	cfg.visitRPO(new Rewrite());
    }

    private static class Interval {
	public final int reg;
	public int start;
	public int end;

	public Interval(int reg, int start, int end) {
            this.reg = reg;
            this.start = start;
            this.end = end;
	}
	public String toString() {
	    return " " + reg + ": <" + start + "," + end+">";
	}
    }

    private static class Allocation {
	public final int temp;
	public final int reg;

	public Allocation(int temp, int reg) {
            this.temp = temp;
            this.reg = reg;
        }
	public String toString() {
	    return "[t" + temp + " -> " + reg + "]";
	}
    }

    private class Tentative {
	public int end;
	public Allocation allocation;
	public Tentative next = null;
        private int spillPos = 32;

	public Tentative(Allocation a, int e) {
	    allocation = a;
	    end = e;
	}
	public Tentative(Allocation allocation, int end, Tentative next) {
	    this.allocation = allocation;
	    this.end = end;
	    this.next = next;
        }
	public Tentative add(Allocation a, int e) {
	    if (e < end)
		return new Tentative(a, e, this);
	    else {
		if (next != null)
		    next = next.add(a, e);
		else
		    next = new Tentative(a, e);
            }
	    return this;
	}
	public Tentative expire(int pos) {
	    if (pos >= end) {
		if (allocation.reg > 1 &&  allocation.reg < 30)
		    regPool.free(allocation.reg);
		finalAllocations[allocation.temp] = allocation.reg;
		if (next != null)
		    return next.expire(pos);
		else
		    return null;
	    } else return this;
	}
	public Tentative spill(int temp, int e) {
	    if (next != null) {
		next = next.spill(temp, e);
		return this;
	    } else {
		if (end < e) {
		    finalAllocations[temp] = spillPos++;
		    return this;
		} else {
		    finalAllocations[allocation.temp] = spillPos++;
		    allocation = new Allocation(temp, allocation.reg);
		    end = e;
		    return this;
		}
	    }
	}
    }

    private class LiveIntervals extends CFG.EmptyVisitor {
	int instructionNr = 0;

	public void pre(CFG.Node node) {
	    BitSet live = liveness.liveIn(node);
	    addPoints(live, instructionNr);

            CFG.InstrVisitor visitor =
                new CFG.InstrVisitor() {
                    public void visit(Instruction instr) {
                        // System.out.println("<"+instructionNr+"> " + instr);
                        // System.out.println("defs " + instr.defs());
                        // System.out.println("uses " + instr.uses());
                        BitSet regs = new BitSet();
                        regs.or(instr.defs());
                        regs.or(instr.uses());
                        addPoints(regs, instructionNr);
                        instructionNr++;
                    }
                };

	    cfg.visitNodeForward(node, visitor);
            live = liveness.liveOut(node);
	    addPoints(live, instructionNr);
	}

        private void addPoints(BitSet bs, int instrNr) {
            for (int i = bs.nextSetBit(0); i >= 0; i = bs.nextSetBit(i + 1)) {
                if (i >= intervals.length) {
                    Interval[] oldIntervals = intervals;
                    intervals =
                        new Interval[Math.max(oldIntervals.length * 2, i + 1)];
                    System.arraycopy(oldIntervals, 0,
                                     intervals, 0,
                                     oldIntervals.length);
                }

                if (intervals[i] == null) {
                    intervals[i] = new Interval(i, instrNr, instrNr);
                } else {
                    Interval itv = intervals[i];
                    itv.start = Math.min(instrNr, itv.start);
                    itv.end   = Math.max(instrNr, itv.end);
                }
            }
        }
    }

    private class Rewrite extends CFG.EmptyVisitor {
	public void pre(CFG.Node node) {
	    cfg.visitNodeForward(node,
                                 new CFG.InstrVisitor() {
                                     public void visit(Instruction instr) {
                                         instr.replaceRegs(finalAllocations);
                                     }
                                 });
	}
    }
}
