/*  miscc -- a compiler for misc
 *    Parser
 *
 *    Matthias Zenger, Michel Schinz, Philippe Altherr
 *    $Id: Parser.java,v 1.1 2004/04/02 14:44:28 schinz Exp $
 */

package miscc;

import java.io.InputStream;
import java.util.Vector;
import miscc.Tree.*;


/** This class implements the parser of the compiler.
 */
class Parser extends Scanner {

    Parser(InputStream in) {
        super(in);
    }

    /** Convert a vector to a tree array.
     */
    private Tree[] toArray(Vector trees) {
        return (Tree[])trees.toArray(new Tree[trees.size()]);
    }

    /** Reports an error that a specific token was expected but
     *  a different token found.
     */
    private void error(int expected) {
        error("token of class " + tokenClass(expected));
    }

    /** Reports an error that a specific token was expected but
     *  a different token found.
     */
    private void error(String expected) {
        String message = "syntax error\n" +
            "expected: " + expected + "\n" +
            "found   : " + representation();
        Report.fail(start, message);
        throw new Error();
    }

    /** If the current token corresponds to the token class 'expected'
     *  then skip it, otherwise signal an error.
     */
    private void accept(int expected) {
        if (token == expected)
            nextToken();
        else
            error(expected);
    }

    /** Parse an identifier.
     */
    private String ident() {
        if (token == IDENT) {
            String name = chars;
            nextToken();
            return name;
        } else {
            nextToken();
            return "<error>";
        }
    }

    /** Parse a natural number (positive).
     */
    private int number() {
        if (token == NUMBER) {
            String num = chars;
            nextToken();
            try {
                return Integer.parseInt(num);
            } catch (NumberFormatException e) {
                error("natural number");
                return 0;
            }
        } else {
            nextToken();
            return 0;
        }
    }

    /** Parse the input and return an abstract syntax tree.
     */
    public Tree parse() {
        Tree result = program();
        accept(EOF);
        return result;
    }

    /** Program        = { Declaration ";" } Body
     */
    private Tree program() {
        int pos = start;
        Vector defs = new Vector();
        while (token == DEF) {
            defs.add(declaration());
            accept(SEMICOLON);
        }
        return new Program(pos, toArray(defs), body());
    }

    /** Declaration    = "def" ident "(" Formals ")" ":" Type "=" Expression
     */
    private Tree declaration() {
        int pos = start;
        accept(DEF);
        String name = ident();
        accept(LPAREN);
        Tree[] params = formals();
        accept(RPAREN);
        accept(COLON);
        Tree restpe = type();
        accept(EQUAL);
        return new FunDef(pos, name, params, restpe, expression());
    }

    /** Body           = { Statement ";" } [ Expression ]
     */
    private Body body() {
        int pos = start;
        Vector stats = new Vector();
        Tree expr = new UnitLit(pos);
        while ((token != EOF) && (token != RBRACE)) {
            if ((token == VAR) || (token == WHILE)) {
                stats.add(statement());
                accept(SEMICOLON);
            } else {
                Tree tree = expression();
                if (token == SEMICOLON) {
                    stats.add(new Exec(tree.pos, tree));
                    nextToken();
                } else {
                    expr = tree;
                    break;
                }
            }
        }
        return new Body(pos, toArray(stats), expr);
    }

    /** Formal         = ident ":" Type
     */
    private Formal formal() {
        int pos = start;
        String name = ident();
        accept(COLON);
        return new Formal(pos, name, type());
    }

    /** Formals        = [ Formal { "," Formal } ]
     */
    private Tree[] formals() {
        Vector args = new Vector();
        if (token == IDENT)
            while (true) {
                args.add(formal());
                if (token == COMMA)
                    nextToken();
                else
                    break;
            }
        return toArray(args);
    }

    /** Type           = "Unit"
     *                 | "Any"
     *                 | "Int"
     *                 | "List" "[" Type "]"
     *                 | "(" Types ")" "=>" Type
     */
    private Tree type() {
        int pos = start;
        if (token == UNIT) {
            nextToken();
            return new BasicType(pos, Tree.UNIT);
        } else if (token == ANY) {
            nextToken();
            return new BasicType(pos, Tree.ANY);
        } else if (token == INT) {
            nextToken();
            return new BasicType(pos, Tree.INT);
        } else if (token == LIST) {
            nextToken();
            accept(LBRACKET);
            Tree elemtpe = type();
            accept(RBRACKET);
            return new ListType(pos, elemtpe);
        } else {
            accept(LPAREN);
            Tree[] argtpes = types();
            accept(RPAREN);
            accept(ARROW);
            return new FunType(pos, argtpes, type());
        }
    }

    /** Types          = [ Type { "," Type } ]
     */
    private Tree[] types() {
        Vector tpes = new Vector();
        if ((token == UNIT) ||
            (token == ANY) ||
            (token == INT) ||
            (token == LIST) ||
            (token == LPAREN))
            while (true) {
                tpes.add(type());
                if (token == COMMA)
                    nextToken();
                else
                    break;
            }
        return toArray(tpes);
    }

    /*  Statement      = "var" Formal "=" Expression
     *                 | "while" "(" Expression ")" Expression
     */
    private Tree statement() {
        int pos = start;
        if (token == VAR) {
            nextToken();
            Formal formal = formal();
            accept(EQUAL);
            return new VarDef(pos, formal, expression());
        } else {
            accept(WHILE);
            accept(LPAREN);
            Tree cond = expression();
            accept(RPAREN);
            return new While(pos, cond, expression());
        }
    }

    /*  Expression     = "if" "(" Expression ")" Expression [ "else" Expression ]
     *                 | ident "=" Expression
     *                 | OrExpression
     */
    private Tree expression() {
        int pos = start;
        if (token == IF) {
            nextToken();
            accept(LPAREN);
            Tree cond = expression();
            accept(RPAREN);
            Tree thenpart = expression();
            if (token == ELSE) {
                nextToken();
                return new If(pos, cond, thenpart, expression());
            } else
                return new If(pos, cond, thenpart, new UnitLit(pos));
        } else if (token == IDENT) {
            Tree expr = orExpression();
            if ((token == EQUAL) && (expr instanceof Ident)) {
                pos = start;
                nextToken();
                return new Assign(pos, ((Ident)expr).name, expression());
            } else
                return expr;
        } else
            return orExpression();
    }

    /** OrExpression   = AndExpression
     *                 | OrExpression "|" AndExpression
     */
    private Tree orExpression() {
        Tree expr = andExpression();
        while (token == OR) {
            int pos = start;
            nextToken();
            Tree notExpr = new Operation(pos, Tree.NOT, expr);
            Tree notAndExpr = new Operation(pos, Tree.NOT, andExpression());
            Tree conj = new Operation(pos, Tree.AND, notExpr, notAndExpr);
            expr = new Operation(pos, Tree.NOT, conj);
        }
        return expr;
    }

    /** AndExpression  = CmpExpression
     *                 | AndExpression "&" CmpExpression
     */
    private Tree andExpression() {
        Tree expr = cmpExpression();
        while (token == AND) {
            int pos = start;
            nextToken();
            expr = new Operation(pos, Tree.AND, expr, cmpExpression());
        }
        return expr;
    }

    /** CmpExpression  = [ ListExpression CompOp ] ListExpression
     *  CompOp         = "=="
     *                 | "!="
     *                 | "<"
     *                 | ">"
     *                 | "<="
     *                 | ">="
     */
    private Tree cmpExpression() {
        Tree expr = listExpression();
        int pos = start;
        switch (token) {
            case EQ:
                nextToken();
                return new Operation(pos, Tree.EQ, expr, listExpression());
            case NE:
                nextToken();
                return new Operation(pos, Tree.NE, expr, listExpression());
            case LT:
                nextToken();
                return new Operation(pos, Tree.LT, expr, listExpression());
            case LE:
                nextToken();
                return new Operation(pos, Tree.LE, expr, listExpression());
            case GT:
                nextToken();
                return new Operation(pos, Tree.GT, expr, listExpression());
            case GE:
                nextToken();
                return new Operation(pos, Tree.GE, expr, listExpression());
            default:
                return expr;
        }
    }

    /** ListExpression = SumExpression
     *                 | SumExpression "::" ListExpression
     */
    private Tree listExpression() {
        Tree expr = sumExpression();
        if (token == CONS) {
            int pos = start;
            nextToken();
            return new Operation(pos, Tree.CONS, expr, listExpression());
        } else
            return expr;
    }

    /** SumExpression  = Term
     *                 | SumExpression SumOp Term
     *  SumOp          = "+" | "-"
     */
    private Tree sumExpression() {
        Tree term = term();
        while (true) {
            int pos = start;
            switch (token) {
                case ADD:
                    nextToken();
                    term = new Operation(pos, Tree.ADD, term, term());
                    break;
                case SUB:
                    nextToken();
                    term = new Operation(pos, Tree.SUB, term, term());
                    break;
                default:
                    return term;
            }
        }
    }

    /** Term           = [ Term ProdOp ] [ NegateOp ] Factor
     *  NegateOp       = "-" | "!"
     *  ProdOp         = "*" | "/" | "%"
     */
    private Tree term() {
        Tree term = signedFactor();
        while (true) {
            int pos = start;
            switch (token) {
                case MUL:
                    nextToken();
                    term = new Operation(pos, Tree.MUL, term, signedFactor());
                    break;
                case DIV:
                    nextToken();
                    term = new Operation(pos, Tree.DIV, term, signedFactor());
                    break;
                case MOD:
                    nextToken();
                    term = new Operation(pos, Tree.MOD, term, signedFactor());
                    break;
                default:
                    return term;
            }
        }
    }

    private Tree signedFactor() {
        int pos = start;
        if (token == SUB) {
            nextToken();
            return new Operation(pos, Tree.SUB, new IntLit(pos, 0), appliedFactor());
        } else if (token == NOT) {
            nextToken();
            return new Operation(pos, Tree.NOT, appliedFactor());
        } else
            return appliedFactor();
    }

    private Tree appliedFactor() {
        Tree factor = factor();
        while (token == LPAREN)
            factor = new Apply(start, factor, params());
        return factor;
    }

    /** Factor         = ident
     *                 | number
     *                 | string
     *                 | "true"
     *                 | "false"
     *                 | "(" [ Expression ] ")"
     *                 | "{" Body "}"
     *                 | PredefFun Params
     *                 | "List" Params
     *                 | Factor Params
     *  PredefFun      = "head"
     *                 | "tail"
     *                 | "isEmpty"
     */
    private Tree factor() {
        int pos = start;
        switch (token) {
            case IDENT:
                return new Ident(pos, ident());
            case NUMBER:
                return new IntLit(pos, number());
            case STRING:
                String str = chars;
                Tree expr = new NilLit(pos);
                for (int i = str.length() - 1; i >= 0; i--)
                    expr = new Operation(pos, Tree.CONS,
                        new IntLit(pos, str.charAt(i)), expr);
                nextToken();
                return expr;
            case TRUE:
                nextToken();
                return new IntLit(pos, 1);
            case FALSE:
                nextToken();
                return new IntLit(pos, 0);
            case LPAREN:
                accept(LPAREN);
                Tree nested = (token != RPAREN) ? expression()
                                                : new UnitLit(pos);
                accept(RPAREN);
                return nested;
            case LBRACE:
                accept(LBRACE);
                Tree body = body();
                accept(RBRACE);
                return body;
            case HEAD:
                nextToken();
                return new Listop(pos, Tree.HEAD, params());
            case TAIL:
                nextToken();
                return new Listop(pos, Tree.TAIL, params());
            case IS_EMPTY:
                nextToken();
                return new Listop(pos, Tree.ISEMPTY, params());
            case LIST:
                nextToken();
                Tree[] elems = params();
                Tree res = new NilLit(pos);
                for (int i = elems.length - 1; i >= 0; i--)
                    res = new Operation(pos, Tree.CONS, elems[i], res);
                return res;
            default:
                error("factor");
                nextToken();
                return new Bad(pos);
        }
    }

    /** Params         = "(" Expressions ")"
     *  Expressions    = [ Expression { "," Expression } ]
     */
    Tree[] params() {
        accept(LPAREN);
        Vector exprs = new Vector();
        if (token != RPAREN)
            while (true) {
                exprs.add(expression());
                if (token == COMMA)
                    nextToken();
                else
                    break;
            }
        accept(RPAREN);
        return toArray(exprs);
    }
}
