package miscc;

import java.util.Map;
import java.util.HashMap;

public interface RISC {

    //########################################################################
    // !!! I/O instructions

    /** <code><b>RD a</b><br>R.a = </code>input*/
    public static final int RD      = 50;
    /** <code><b>WRD c</b><br></code>prints <code>R.c</code>*/
    public static final int WRD     = 51;
    /** <code><b>WRH c</b><br></code>prints <code>R.c</code>*/
    public static final int WRH     = 52;
    /** <code><b>WRL</b><br></code>prints a newline*/
    public static final int WRL     = 53;

    //########################################################################
    // Opcodes - instructions with registers

    /**
     * <p>Opcode: integer addition.</p>
     * <b><code>ADD a b c</code></b><br>
     *    <code>R.a = R.b + R.c</code>
     */
    public static final int ADD  =  0;

    /**
     * <p>Opcode: integer subtraction.</p>
     * <b><code>SUB a b c</code></b><br>
     *    <code>R.a = R.b - R.c</code>
     */
    public static final int SUB  =  1;

    /**
     * <p>Opcode: integer multiplication.</p>
     * <b><code>MUL a b c</code></b><br>
     *    <code>R.a = R.b * R.c</code>
     */
    public static final int MUL  =  2;

    /**
     * <p>Opcode: integer division.</p>
     * <b><code>DIV a b c</code></b><br>
     *    <code>R.a = R.b / R.c</code>
     */
    public static final int DIV  =  3;

    /**
     * <p>Opcode: integer modulo.</p>
     * <b><code>MOD a b c</code></b><br>
     *    <code>R.a = R.b % R.c</code>
     */
    public static final int MOD  =  4;

    /**
     * <p>Opcode: integer comparison.</p>
     * <b><code>CMP a b c</code></b><br>
     *    <code>R.a = &lt;a value with the same sign as (R.b - R.c)
     *    but with a possibly different magnitude&gt;</code>
     */
    public static final int CMP  =  5;

    /**
     * <p>Opcode: logical or.</p>
     * <b><code>OR a b c</code></b><br>
     *    <code>R.a = R.b | R.c</code>
     */
    public static final int OR   =  8;

    /**
     * <p>Opcode: logical and.</p>
     * <b><code>AND a b c</code></b><br>
     *    <code>R.a = R.b & R.c</code>
     */
    public static final int AND  =  9;

    /**
     * <p>Opcode: logical bic.</p>
     * <b><code>BIC a b c</code></b><br>
     *    <code>R.a = R.b & ~R.c</code>
     */
    public static final int BIC  = 10;

    /**
     * <p>Opcode: logical xor.</p>
     * <b><code>XOR a b c</code></b><br>
     *    <code>R.a = R.b ^ R.c</code>
     */
    public static final int XOR  = 11;

    /**
     * <p>Opcode: logical shift.</p>
     * <b><code>LSH a b c</code></b><br>
     *    <code>R.a = (R.c &gt; 0) ?
     *      (R.b &lt;&lt; R.c) : (R.b &gt;&gt;&gt; -R.c)</code>
     */
    public static final int LSH  = 12;

    /**
     * <p>Opcode: arithmetic shift.</p>
     * <b><code>ASH a b c</code></b><br>
     *    <code>R.a = (R.c &gt; 0) ?
     *      (R.b &lt;&lt; R.c) : (R.b &gt;&gt; -R.c)</code>
     */
    public static final int ASH  = 13;

    /**
     * <p>Opcode: bound check.</p>
     * <b><code>CHK a c</code></b><br>
     *    raise an error if not <code>(0 &lt;= R.a < R.c)</code>
     */
    public static final int CHK  = 14;

    //########################################################################
    // Opcodes - instructions with register and signed immediate

    /**
     * <p>Opcode: integer addition with signed immediate.</p>
     * <b><code>ADDI a b ic</code></b><br>
     *    <code>R.a = R.b + ic</code>
     */
    public static final int ADDI    = 16;

    /**
     * <p>Opcode: integer subtraction with signed immediate.</p>
     * <b><code>SUBI a b ic</code></b><br>
     *    <code>R.a = R.b - ic</code>
     */
    public static final int SUBI    = 17;

    /**
     * <p>Opcode: integer multiplication with signed immediate.</p>
     * <b><code>MULI a b ic</code></b><br>
     *    <code>R.a = R.b * ic</code>
     */
    public static final int MULI    = 18;

    /**
     * <p>Opcode: integer division with signed immediate.</p>
     * <b><code>DIVI a b ic</code></b><br>
     *    <code>R.a = R.b / ic</code>
     */
    public static final int DIVI    = 19;

    /**
     * <p>Opcode: integer modulo with signed immediate.</p>
     * <b><code>MODI a b ic</code></b><br>
     *    <code>R.a = R.b % ic</code>
     */
    public static final int MODI    = 20;

    /**
     * <p>Opcode: integer comparison with signed immediate.</p>
     * <b><code>CMPI a b ic</code></b><br>
     *    <code>R.a = &lt;a value with the same sign as (R.b - ic)
     *    but with a possibly different magnitude&gt;</code>
     */
    public static final int CMPI    = 21;

    /**
     * <p>Opcode: logical or with signed immediate.</p>
     * <b><code>ORI a b ic</code></b><br>
     *    <code>R.a = R.b | ic</code>
     */
    public static final int ORI     = 24;

    /**
     * <p>Opcode: logical and with signed immediate.</p>
     * <b><code>ANDI a b ic</code></b><br>
     *    <code>R.a = R.b & ic</code>
     */
    public static final int ANDI    = 25;

    /**
     * <p>Opcode: logical bic with signed immediate.</p>
     * <b><code>BICI a b ic</code></b><br>
     *    <code>R.a = R.b & ~ic</code>
     */
    public static final int BICI    = 26;

    /**
     * <p>Opcode: logical xor with signed immediate.</p>
     * <b><code>XORI a b ic</code></b><br>
     *    <code>R.a = R.b ^ ic</code>
     */
    public static final int XORI    = 27;

    /**
     * <p>Opcode: logical shift with signed immediate.</p>
     * <b><code>LSHI a b ic</code></b><br>
     *    <code>R.a = (ic &gt; 0) ?
     *      (R.b &lt;&lt; ic) : (R.b &gt;&gt;&gt; -ic)</code>
     */
    public static final int LSHI    = 28;

    /**
     * <p>Opcode: arithmetic shift with signed immediate.</p>
     * <b><code>ASHI a b ic</code></b><br>
     *    <code>R.a = (ic &gt; 0) ?
     *      (R.b &lt;&lt; ic) : (R.b &gt;&gt; -ic)</code>
     */
    public static final int ASHI    = 29;

    /**
     * <p>Opcode: bound check with signed immediate.</p>
     * <b><code>CHKI a ic</code></b><br>
     *    raise an error if not <code>(0 &lt;= R.a < ic)</code>
     */
    public static final int CHKI    = 30;

    //########################################################################
    // Opcodes - instructions with register and unsigned immediate

    /**
     * <p>Opcode: integer addition with unsigned immediate.</p>
     * <b><code>ADDIU a b uc</code></b><br>
     *    <code>R.a = R.b + uc</code>
     */
    public static final int ADDIU   = 54;

    /**
     * <p>Opcode: integer subtraction with unsigned immediate.</p>
     * <b><code>SUBIU a b uc</code></b><br>
     *    <code>R.a = R.b - uc</code>
     */
    public static final int SUBIU   = 55;

    /**
     * <p>Opcode: integer multiplication with unsigned immediate.</p>
     * <b><code>MULIU a b uc</code></b><br>
     *    <code>R.a = R.b * uc</code>
     */
    public static final int MULIU   = 56;

    /**
     * <p>Opcode: integer division with unsigned immediate.</p>
     * <b><code>DIVIU a b uc</code></b><br>
     *    <code>R.a = R.b / uc</code>
     */
    public static final int DIVIU   = 57;

    /**
     * <p>Opcode: integer modulo with unsigned immediate.</p>
     * <b><code>MODIU a b uc</code></b><br>
     *    <code>R.a = R.b % uc</code>
     */
    public static final int MODIU   = 58;

    /**
     * <p>Opcode: integer comparison with unsigned immediate.</p>
     * <b><code>CMPIU a b uc</code></b><br>
     *    <code>R.a = &lt;a value with the same sign as (R.b - uc)
     *    but with a possibly different magnitude&gt;</code>
     */
    public static final int CMPIU   = 59;

    /**
     * <p>Opcode: logical or with unsigned immediate.</p>
     * <b><code>ORIU a b uc</code></b><br>
     *    <code>R.a = R.b | uc</code>
     */
    public static final int ORIU    = 60;

    /**
     * <p>Opcode: logical and with unsigned immediate.</p>
     * <b><code>ANDIU a b uc</code></b><br>
     *    <code>R.a = R.b & uc</code>
     */
    public static final int ANDIU   = 61;

    /**
     * <p>Opcode: logical bic with unsigned immediate.</p>
     * <b><code>BICIU a b uc</code></b><br>
     *    <code>R.a = R.b & ~uc</code>
     */
    public static final int BICIU   = 62;

    /**
     * <p>Opcode: logical xor with unsigned immediate.</p>
     * <b><code>XORIU a b uc</code></b><br>
     *    <code>R.a = R.b ^ uc</code>
     */
    public static final int XORIU   = 63;

    /**
     * <p>Opcode: bound check with unsigned immediate.</p>
     * <b><code>CHKIU a uc</code></b><br>
     *    raise an error if not <code>(0 &lt;= R.a < uc)</code>
     */
    public static final int CHKIU   = 39;

    //########################################################################
    // Opcodes - load/store instructions

    /**
     * <p>Opcode: load word from memory. The address <code>R.b</code>
     * must be aligned on a word boundary.</p>
     *
     * <b><code>LDW a b ic</code></b><br>
     *    <code>R.a = &lt;word at address R.b + ic&gt;</code>
     */
    public static final int LDW  = 32;

    /**
     * <p>Opcode: load byte from memory.</p>
     *
     * <b><code>LDB a b ic</code></b><br>
     *    <code>R.a = &lt;byte at address R.b + ic&gt;</code>
     */
    public static final int LDB  = 33;

    /**
     * <p>Opcode: pop word from stack. The address <code>R.b</code> must
     * be aligned on a word boundary.</p>
     *
     * <b><code>POP a b ic</code></b><br>
     *    <code>R.a = &lt;word at address R.b&gt;</code><br>
     *    <code>R.b = R.b + ic</code>
     */
    public static final int POP  = 34;

    /**
     * <p>Opcode: store word into memory. The address <code>R.b</code>
     * must be aligned on a word boundary.</p>
     *
     * <b><code>STW a b ic</code></b><br>
     *    <code>&lt;word at address R.b + ic&gt; = R.a</code>
     */
    public static final int STW  = 36;

    /**
     * <p>Opcode: store byte into memory.</p>
     *
     * <b><code>STB a b ic</code></b><br>
     *    <code>&lt;byte at address R.b + ic&gt; = (byte)R.a</code>
     */
    public static final int STB  = 37;

    /**
     * <p>Opcode: push word on stack. The address <code>R.b</code> must
     * be aligned on a word boundary.</p>
     *
     * <b><code>PSH a b ic</code></b><br>
     *    <code>R.b = R.b - ic</code><br>
     *    <code>&lt;word at address R.b&gt; = R.a</code>
     */
    public static final int PSH  = 38;

    //########################################################################
    // Opcodes - control instructions

    /**
     * <p>Opcode: branch if equal.</p>
     * <b><code>BEQ a oc</code></b><br>
     *    branch to <code>(PC + 4*oc)</code> if <code>(R.a == 0)</code>
     */
    public static final int BEQ  = 40;

    /**
     * <p>Opcode: branch if not equal.</p>
     * <b><code>BNE a oc</code></b><br>
     *    branch to <code>(PC + 4*oc)</code> if <code>(R.a != 0)</code>
     */
    public static final int BNE  = 41;

    /**
     * <p>Opcode: branch if less than.</p>
     * <b><code>BLT a oc</code></b><br>
     *    branch to <code>(PC + 4*oc)</code> if <code>(R.a < 0)</code>
     */
    public static final int BLT  = 42;

    /**
     * <p>Opcode: branch if greater or equal.</p>
     * <b><code>BGE a oc</code></b><br>
     *    branch to <code>(PC + 4*oc)</code> if <code>(R.a >= 0)</code>
     */
    public static final int BGE  = 43;

    /**
     * <p>Opcode: branch if equal less or equal.</p>
     * <b><code>BLE a oc</code></b><br>
     *    branch to <code>(PC + 4*oc)</code> if <code>(R.a <= 0)</code>
     */
    public static final int BLE  = 44;

    /**
     * <p>Opcode: branch if greater than.</p>
     * <b><code>BGT a oc</code></b><br>
     *    branch to <code>(PC + 4*oc)</code> if <code>(R.a > 0)</code>
     */
    public static final int BGT  = 45;

    /**
     * <p>Opcode: branch to subroutine.</p>
     * <b><code>BSR oc</code></b><br>
     *    <code>R.31 = PC + 4</code><br>
     *    branch to <code>(PC + 4*oc)</code>
     */
    public static final int BSR  = 46;

    /**
     * <p>Opcode: jump to subroutine.</p>
     * <b><code>JSR lc</code></b><br>
     *    <code>R.31 = PC + 4</code><br>
     *    branch to <code>(4*lc)</code>
     */
    public static final int JSR  = 48;

    /**
     * <p>Opcode: jump to return address.</p>
     * <b><code>RET c</code></b><br>
     *    jump to <code>R.c</code>
     */
    public static final int RET  = 49;

    //########################################################################
    // Opcodes - miscellaneous instructions

    /**
     * <p>Opcode: stop execution and return to debugger.</p>
     * <b><code>BREAK</code></b><br>
     *    stop execution and return to debugger
     */

    public static final int BREAK   = 6;


    /**
     * <p>Opcode: invoke a system function.</p>
     * <b><code>SYSCALL a b uc</b></code><br>
     *    invoke system function <code>uc</code> with registers
     *    <code>R.a</code> and <code>R.b</code>
     */
    public static final int SYSCALL = 7;


    //########################################################################
    // Symbolic Instructions

    /**
     * <p>Opcode: Name a place in the code..</p>
     * <b><code>LABEL n</code></b><br>
     */
    public static final int LABEL   = 64;

    /**
     * <p>Symbolic Opcode: load address.</p>
     *
     * <b><code>LDA a label</code></b><br>
     *    <code>R.a = address of label</code>
     * Syntactic sugar for:<br/>
     *    <code>ORI a 0 &lt;address of label&gt;</code>
     *    <code>R.a = R.0 or &lt;address of label&gt;</code>
     */
    public static final int LDA  = 65;

    /**
     * <p>Symbolic Opcode: call address.</p>
     *
     * <b><code>CALL R.a</code></b><br/>
     *    <code>R.31 = PC + 4</code><br/>
     *    jump to <code>R.a</code>
     *
     * Syntactic sugar for:<br/>
     *    <code>ORI R.31 0 &lt;address of Label&gt;</code><br/>
     *    <code>RET R.a</code>
     */
    public static final int CALL  = 66;

    //########################################################################
    // System calls - IO read

    /**
     * <p>System call: read one character.</p>
     * <b><code>SYSCALL a 0 SYS_IO_RD_CHR</b></code><br>
     *    <code>R.a = &lt;Unicode of read character or -1 if EOF&gt;</code>
     */
    public static final int SYS_IO_RD_CHR = 1;

    /**
     * <p>System call: read an integer.</p>
     * <b><code>SYSCALL a 0 SYS_IO_RD_INT</b></code><br>
     *    <code>R.a = &lt;value of read integer&gt;</code>
     */
    public static final int SYS_IO_RD_INT = 2;

    /**
     * <p>System call: read a binary number.</p>
     * <b><code>SYSCALL a 0 SYS_IO_RD_BIN</b></code><br>
     *    <code>R.a = &lt;value of read binary number&gt;</code>
     */
    public static final int SYS_IO_RD_BIN = 3;

    /**
     * <p>System call: read a decimal number.</p>
     * <b><code>SYSCALL a 0 SYS_IO_RD_DEC</b></code><br>
     *    <code>R.a = &lt;value of read decimal number&gt;</code>
     */
    public static final int SYS_IO_RD_DEC = 4;

    /**
     * <p>System call: read a hexadecimal number.</p>
     * <b><code>SYSCALL a 0 SYS_IO_RD_HEX</b></code><br>
     *    <code>R.a = &lt;value of read hexadecimal number&gt;</code>
     */
    public static final int SYS_IO_RD_HEX = 5;

    //########################################################################
    // System calls - IO write

    /**
     * <p>System call: write one character.</p>
     * <b><code>SYSCALL a 0 SYS_IO_WR_CHR</b></code><br>
     *    <code>&lt;write character with Unicode R.a&gt;</code>
     */
    public static final int SYS_IO_WR_CHR = 6;

    /**
     * <p>System call: write an integer.</p>
     * <b><code>SYSCALL a b SYS_IO_WR_INT</b></code><br>
     *    <code>&lt;write signed value R.a in decimal format and
     *          space padded to width R.b&gt;</code>
     */
    public static final int SYS_IO_WR_INT = 7;

    /**
     * <p>System call: write a binary number.</p>
     * <b><code>SYSCALL a b SYS_IO_WR_BIN</b></code><br>
     *    <code>&lt;write unsigned value R.a in binary format and
     *          zero padded to width R.b&gt;</code>
     */
    public static final int SYS_IO_WR_BIN = 8;

    /**
     * <p>System call: write a decimal number.</p>
     * <b><code>SYSCALL a b SYS_IO_WR_DEC</b></code><br>
     *    <code>&lt;write unsigned value R.a in decimal format and
     *          zero padded to width R.b&gt;</code>
     */
    public static final int SYS_IO_WR_DEC = 9;

    /**
     * <p>System call: write a hexadecimal number.</p>
     * <b><code>SYSCALL a b SYS_IO_WR_HEX</b></code><br>
     *    <code>&lt;write unsigned value R.a in hexadecimal format and
     *          zero padded to width R.b&gt;</code>
     */
    public static final int SYS_IO_WR_HEX = 10;

    /**
     * <p>System call: write a line separator.</p>
     * <b><code>SYSCALL 0 0 SYS_IO_WR_LN</b></code><br>
     *    <code>&lt;write a line separator&gt;</code>
     */
    public static final int SYS_IO_WR_LN = 17;

    //########################################################################
    // System calls - IO miscellaneous

    /**
     * <p>System call: flush the output stream.</p>
     * <b><code>SYSCALL 0 0 SYS_IO_FLUSH</b></code><br>
     *    <code>&lt;flush the output stream&gt;</code>
     */
    public static final int SYS_IO_FLUSH = 15;

    //########################################################################
    // System calls - garbage collector

    /**
     * <p>System call: initialize the garbage collector.</p>
     *
     * <p>The garbage collector is initialized with an empty heap that
     * starts at address <code>R.a</cod> and with a maximum size of
     * <code>sz = R.b & 0x07FFFFFF</code> words. If <code>sz</code> is
     * zero then the heap extends to the end of the memory.</p>
     *
     * <p>The value <code>sp = R.b &gt;&gt;&gt; 27</code> determines
     * whether there is a stack and which register is the stack
     * pointer: if <code>sp</code> is non-zero, the garbage collector
     * assumes that there is a stack and that <code>R.sp</code> is the
     * stack pointer register. It is assumed that the stack starts at
     * the end of the memory and grows downwards.<p>
     *
     * <p>During a garbage collection, the garbage collector frees all
     * the memory blocks which are not referenced by any live
     * pointer. A live pointer (resp. value) is either a root pointer
     * (resp. value) or a pointer (resp. value) contained in a block
     * referenced by a live pointer. Note that as there is no way to
     * distinguish between pointers and non-pointer values, all values
     * are regarded as pointers. A root value is one that is contained
     * in a register, in the memory below the heap or in the stack. If
     * there is no stack, all the values contained in the memory above
     * the heap are also root values.<p>
     *
     * <b><code>SYSCALL a b SYS_GC_INIT</b></code><br>
     *    <code>&lt;initialize the garbage collector&gt;</code>
     */
    public static final int SYS_GC_INIT = 11;

    /**
     * <p>System call: allocate a memory block from the heap.</p>
     *
     * <p>Allocates a memory block of, at least, <code>R.b</code>
     * bytes from the heap and returns its address in
     * <code>R.a</code>. The allocated memory is zero-initialized and
     * its address is guaranteed to be aligned on a word boundary. If
     * there is not enough free memory, a garbage collection is
     * triggered and if this doesn't free enough memory, an error is
     * raised and the execution stopped.<p>
     *
     * <b><code>SYSCALL a b SYS_GC_ALLOC</b></code><br>
     *    <code>R.a = &lt;address of the newly allocated and
     *          zero-initialized memory block of R.b bytes&gt;</code>
     */
    public static final int SYS_GC_ALLOC = 12;

    //########################################################################
    // System calls - miscellaneous

    /**
     * <p>System call: get memory size.</p>
     * <b><code>SYSCALL a 0 SYS_GET_TOTAL_MEM_SIZE</b></code><br>
     *    <code>R.a = &lt;memory size in bytes&gt;</code>
     */
    public static final int SYS_GET_TOTAL_MEM_SIZE = 13;

    /**
     * <p>System call: terminate the emulation.</p>
     * <b><code>SYSCALL a 0 SYS_EXIT</b></code><br>
     *    <code>&lt;terminates the emulation with status code R.a&gt;</code>
     */
    public static final int SYS_EXIT = 19;

    //########################################################################
    // System calls - debug

    // !!!
    public static final int SYS_DB_WR_REG = 16;

    // !!!
    public static final int SYS_DB_DUMP = 18;

    //########################################################################
    // Instruction formats

    public static final int OpRaRbRc =  1;
    public static final int OpRaRbIc =  2;
    public static final int OpRaRbUc =  3;

    public static final int OpRa__Rc =  4;
    public static final int OpRa__Ic =  5;
    public static final int OpRa__Uc =  6;
    public static final int OpRa__Oc =  7;
    public static final int OpRa____ =  8;

    public static final int Op____Rc =  9;
    public static final int Op____Oc = 10;
    public static final int Op____Lc = 11;
    public static final int Op______ = 12;

    //########################################################################
    // !!!

    /** Total number of instructions */
    public static final int opcode_count = 67;

    /** String representation of instructions */
    public static final String[] opcode_mnemonic = new String[opcode_count];

    /** Format of instructions */
    public static final int[] opcode_format = new int[opcode_count];

    //########################################################################
    // Registers

    /** Register */
    public final int
        R0  =  0,
        R1  =  1,
        R2  =  2,
        R3  =  3,
        R4  =  4,
        R5  =  5,
        R6  =  6,
        R7  =  7,
        R8  =  8,
        R9  =  9,
        R10 = 10,
        R11 = 11,
        R12 = 12,
        R13 = 13,
        R14 = 14,
        R15 = 15,
        R16 = 16,
        R17 = 17,
        R18 = 18,
        R19 = 19,
        R20 = 20,
        R21 = 21,
        R22 = 22,
        R23 = 23,
        R24 = 24,
        R25 = 25,
        R26 = 26,
        R27 = 27,
        R28 = 28,
        R29 = 29,
        R30 = 30,
        R31 = 31;

    /** Register: zero */
    public static final int ZERO =  R0;

    /** Register: function result */
    public static final int RES  = R1;

    /** Register: stack pointer */
    public final int SP   = R30;

    /** Register: return address */
    public final int LNK  = R31;

    //########################################################################
    // !!!

    /** Total number of registers */
    public static final int register_count = 32;

    /** String representation of registers */
    public static final String[] register_name = new String[register_count];


    //########################################################################
    // !!!

    /** Data initialization status */
    public static final boolean initialized = Helper.initialize();

    //########################################################################

    // !!!
    public static final int WORD_SIZE = 4;

    //########################################################################

    public class Helper {

	public static String pad(int width, String string, char pad) {
	    return string + pad(width - string.length(), pad);
	}

	public static String pad(int width, char pad, String string) {
	    return pad(width - string.length(), pad) + string;
	}

	public static String pad(int width, char pad) {
	    if (width <= 0) return "";
	    StringBuffer buffer = new StringBuffer(width);
	    for (int i = 0; i < width; i++) buffer.append(pad);
	    return buffer.toString();
	}

	public static String toHexString(int value) {
	    return Integer.toHexString(value).toUpperCase();
	}

	public static String toHexString(int value, int width) {
	    return pad(width, '0', toHexString(value));
	}

        // !!!
        public static String getAddressString(int address) {
            return "0x" + toHexString(address, 8);
        }

        public static int op(int code) { return (code             ) >>> 26; }
        public static int ra(int code) { return (code & 0x03E00000) >>> 21; }
        public static int rb(int code) { return (code & 0x001F0000) >>> 16; }
        public static int rc(int code) { return (code & 0x0000001F)       ; }
        public static int ic(int code) { return (code <<        16) >>  16; }
        public static int uc(int code) { return (code & 0x0000FFFF)       ; }
        public static int oc(int code) { return (code <<        16) >>  16; }
        public static int lc(int code) { return (code & 0x03FFFFFF)       ; }

        public static int code(int op, int a, int b, int c) {
            // !!! check : 0 <= op length
            // !!! add checks for a, b and c
            int format = opcode_format[op];
            switch (format) {
            case RISC.OpRaRbRc: return op << 26 | a << 21 | b << 16 | c;
            case RISC.OpRaRbIc: return op << 26 | a << 21 | b << 16 | c&0xFFFF;
            case RISC.OpRaRbUc: return op << 26 | a << 21 | b << 16 | c&0xFFFF;
            case RISC.OpRa__Rc: return op << 26 | a << 21           | c;
            case RISC.OpRa__Ic: return op << 26 | a << 21           | c&0xFFFF;
            case RISC.OpRa__Uc: return op << 26 | a << 21           | c&0xFFFF;
            case RISC.OpRa__Oc: return op << 26 | a << 21          |c&0x1FFFFF;
            case RISC.OpRa____: return op << 26 | a << 21;
            case RISC.Op____Rc: return op << 26                      |c&0xFFFF;
            case RISC.Op____Oc: return op << 26                    |c&0x1FFFFF;
            case RISC.Op____Lc: return op << 26                   |c&0x3FFFFFF;
            case RISC.Op______: return op << 26;
                // !!! java.lang
            default: throw new java.lang.Error("unknown instruction format " + format); // !!! error
            }
        }

        public static Map getMnemonicToOpcodeMap() {
            HashMap map = new HashMap(opcode_count);
            for (int i = 0; i < opcode_count; i++) {
                String mnemonic = opcode_mnemonic[i];
                if (mnemonic != null) map.put(mnemonic, new Integer(i));
            }
            return map;
        }

        public static Map getNameToRegisterMap() {
            HashMap map = new HashMap(register_count);
            for (int i = 0; i < register_count; i++) {
                String name = register_name[i];
                if (name != null) map.put(name, new Integer(i));
            }
            return map;
        }

        public static boolean initialize() {
            for (int i = 0; i < opcode_count; ++i) {
                if (Opcode.codes[i] != null) {
                    opcode_mnemonic[i] = Opcode.codes[i].mnemonic;
                    opcode_format[i] = Opcode.codes[i].format;
                }
            }

            for (int i = 0; i < register_count; i++)
                initRegister(i, "R" + i);

            return true;
        }

        public static void initRegister(int register, String name) {
            register_name[register] = name;
        }
    }

    //########################################################################
}
