package miscc;

import java.util.*;

/**
 * Introduce spilling. This is done by walking over a CFG representing
 * a complete function, and replacing accesses to "temporary"
 * registers with accesses to special spilling locations on the stack.
 */

public class Spiller {
    private final static int firstSpillReg = 28;
    private final static int spillRegsCount = 2;

    public void rewriteSpills(final CFG cfg) {
        // First pass: compute number of spilling locations required.
        final BitSet fakeRegs = new BitSet();
        cfg.visitDFS(new CFG.EmptyVisitor() {
                public void pre(CFG.Node node) {
                    Iterator instrIt = node.code.instructions().iterator();
                    while (instrIt.hasNext()) {
                        Instruction instr = (Instruction)instrIt.next();
                        fakeRegs.or(instr.defs());
                        fakeRegs.or(instr.uses());
                    }
                }
            });
        fakeRegs.clear(0, 32);

        if (fakeRegs.isEmpty())
            return;             // Nothing to spill, code unchanged.

        // Assign spilling location.
        final int spillSize = fakeRegs.cardinality() * RISC.WORD_SIZE;
        final int[] spillIndex = new int[fakeRegs.length() + 1];
        Arrays.fill(spillIndex, -1);
        int currIdx = 0;
        for (int i = fakeRegs.nextSetBit(0);
             i >= 0;
             i = fakeRegs.nextSetBit(i + 1)) {
            spillIndex[i] = currIdx;
            currIdx += RISC.WORD_SIZE;
        }

        // Second pass: rewrite code
        cfg.visitDFS(new CFG.EmptyVisitor() {
                HashMap/*<CFG.Node, Integer>*/ frameSizes = new HashMap();
                { frameSizes.put(cfg.getNodes().get(0), new Integer(0)); }

                public void pre(CFG.Node node) {
                    List/*<Edge>*/ outEdges = cfg.getEdgesFrom(node);
                    boolean entryNode = cfg.getEdgesTo(node).isEmpty();
                    boolean exitNode = outEdges.isEmpty();
                    int frameSize = ((Integer)frameSizes.get(node)).intValue();

                    Code newCode = new Code();

                    if (entryNode) {
                        // Entry node, add spill-region allocation
                        assert frameSize == 0 : frameSize;
                        newCode.emit(RISC.SUBI, RISC.SP, RISC.SP, spillSize);
                    }

                    Iterator instrIt = node.code.instructions().iterator();
                    while (instrIt.hasNext()) {
                        int spillRegIdx = 0;
                        Instruction instr = (Instruction)instrIt.next();
                        BitSet defdRegs = instr.defs();
                        defdRegs.and(fakeRegs);
                        BitSet usedRegs = instr.uses();
                        usedRegs.and(fakeRegs);

                        BitSet allRegs = new BitSet();
                        allRegs.or(defdRegs);
                        allRegs.or(usedRegs);

                        assert allRegs.cardinality() <= spillRegsCount
                            : instr;

                        // Compute register mapping
                        int[] regMap = new int[allRegs.length()];
                        for (int i = 0; i < regMap.length; ++i) {
                            if (allRegs.get(i)) {
                                regMap[i] = firstSpillReg + spillRegIdx;
                                spillRegIdx =
                                    (spillRegIdx + 1) % spillRegsCount;
                            } else
                                regMap[i] = i;
                        }

                        instr.replaceRegs(regMap);

                        // Unspill used fake registers
                        for (int i = usedRegs.nextSetBit(0);
                             i >= 0;
                             i = usedRegs.nextSetBit(i + 1)) {
                            assert spillIndex[i] != -1;
                            int offset = frameSize + spillIndex[i];
                            newCode.emit(RISC.LDW, regMap[i], RISC.SP, offset);
                        }

                        // Add deallocation of spill region, if just
                        // before last instruction.
                        if (exitNode && instr.opcode == Opcode.RET)
                            newCode.emit(RISC.ADDI,
                                         RISC.SP,
                                         RISC.SP,
                                         spillSize);

                        // Rewrite stack accesses, if needed.
                        Instruction newInstr = instr;
                        if (instr.opcode == Opcode.STW
                            || instr.opcode == Opcode.LDW) {
                            OpRaRbIc instr1 = (OpRaRbIc)instr;
                            if (instr1.b == RISC.SP && instr1.c >= frameSize) {
                                newInstr =
                                    Instruction.make(instr1.opcode.code,
                                                     instr1.a,
                                                     instr1.b,
                                                     instr1.c + spillSize);
                            }
                        }
                        newCode.addLast(newInstr);

                        // Record modifications to stack size.
                        frameSize += frameSizeIncrement(instr);

                        // Spill defined fake registers
                        for (int i = defdRegs.nextSetBit(0);
                             i >= 0;
                             i = defdRegs.nextSetBit(i+1)) {
                            assert spillIndex[i] >= 0;
                            int offset = frameSize + spillIndex[i];
                            newCode.emit(RISC.STW, regMap[i], RISC.SP, offset);
                        }
                    }

                    if (!exitNode) {
                        // Take care of out edges
                        Instruction prototype =
                            ((CFG.Edge)outEdges.get(0)).branch;

                        BitSet usedRegs = prototype.uses();
                        usedRegs.and(fakeRegs);

                        int[] regMap = new int[usedRegs.length()];
                        for (int i = 0; i < regMap.length; ++i) {
                            if (usedRegs.get(i)) {
                                regMap[i] = firstSpillReg;
                            } else
                                regMap[i] = i;
                        }

                        for (int i = usedRegs.nextSetBit(0);
                             i >= 0;
                             i = usedRegs.nextSetBit(i + 1)) {
                            assert spillIndex[i] != -1;
                            int offset = frameSize + spillIndex[i];
                            newCode.emit(RISC.LDW,
                                         firstSpillReg,
                                         RISC.SP,
                                         offset);
                        }

                        Iterator outEdgesIt = outEdges.iterator();
                        while (outEdgesIt.hasNext()) {
                            CFG.Edge edge = (CFG.Edge)outEdgesIt.next();
                            edge.branch.replaceRegs(regMap);
                        }
                    }

                    node.code = newCode;
                    Iterator succIt = cfg.successors(node).iterator();
                    while (succIt.hasNext())
                        frameSizes.put(succIt.next(), new Integer(frameSize));
                }

                private int frameSizeIncrement(Instruction instr) {
                    switch (instr.opcode.code) {
                    case RISC.PSH: {
                        PSH instr1 = (PSH)instr;
                        return (instr1.b == RISC.SP) ? instr1.c : 0;
                    }

                    case RISC.POP: {
                        POP instr1 = (POP)instr;
                        return (instr1.b == RISC.SP) ? -instr1.c : 0;
                    }

                    case RISC.ADDI: {
                        ADDI instr1 = (ADDI)instr;
                        return (instr1.a == RISC.SP) ? -instr1.c : 0;
                    }

                    default:
                        assert !instr.defs().get(RISC.SP);
                        return 0;
                    }
                }
            });
    }
}
