/*  miscc -- a compiler for misc
 *    Type representation
 *
 *    Matthias Zenger, Michel Schinz, Philippe Altherr
 *    $Id: Type.java,v 1.1 2004/04/02 14:44:24 schinz Exp $
 */

package miscc;


/** This class provides a representation for types in misc
 */
class Type {

    /** This singleton type is used when there is no type associated
     *  with a tree node.
     */
    static final Type NO_TYPE = new Type();

    /** This type represents an erroneous type (caused by a typing
     *  error).
     */
    static final Type BAD_TYPE = new Type();

    /** The bottom type.
     */
    static final Type BOTTOM_TYPE = new Type();

    /** The top type.
     */
    static final Type ANY_TYPE = new Type();

    /** Type Unit.
     */
    static final Type UNIT_TYPE = new Type();

    /** Type Int.
     */
    static final Type INT_TYPE = new Type();

    /** This method can be used to check if two types are equal.
     */
    boolean isSameType(Type that) {
        return (this == that) ||
               (this == BAD_TYPE) ||
               (that == BAD_TYPE);
    }

    /** This method checks if the provided type is a super type
     *  of the receiver.
     */
    boolean isSubType(Type that) {
        return (this == BAD_TYPE) ||
               (that == BAD_TYPE) ||
               (this == BOTTOM_TYPE) ||
               (that == ANY_TYPE) ||
               isSameType(that);
    }

    /** Return the less specific type of this type and 'that'.
     */
    Type max(Type that) {
        if (isSubType(that))
            return that;
        if (that.isSubType(this))
            return this;
        return BAD_TYPE;
    }

    /** Returns a textual representation of this type.
     */
    public String toString() {
        if (this == NO_TYPE)
            return "<none>";
        if (this == BAD_TYPE)
            return "<bad>";
        if (this == BOTTOM_TYPE)
            return "<bottom>";
        if (this == ANY_TYPE)
            return "Any";
        if (this == UNIT_TYPE)
            return "Unit";
        if (this == INT_TYPE)
            return "Int";
        throw new Error();
    }

    /** Representation for list types.
     */
    static class ListType extends Type {
        Type elemType;

        ListType(Type elemType) {
            this.elemType = elemType;
        }

        boolean isSameType(Type that) {
            return (that == BAD_TYPE) ||
                   ((that instanceof ListType) &&
                    elemType.isSameType(((ListType)that).elemType));
        }

        boolean isSubType(Type that) {
            return (that == BAD_TYPE) ||
                   (that == ANY_TYPE) ||
                   ((that instanceof ListType) &&
                    elemType.isSubType(((ListType)that).elemType));
        }

        public String toString() {
            return "List[" + elemType + "]";
        }
    }

    /** Representation for function types.
     */
    static class FunType extends Type {
        Type[] argTypes;
        Type resType;

        FunType(Type[] argTypes, Type resType) {
            this.argTypes = argTypes;
            this.resType = resType;
        }

        boolean isSameType(Type that) {
            return (that == BAD_TYPE) ||
                   ((that instanceof FunType) &&
                    sameArgTypes((FunType)that) &&
                    resType.isSameType(((FunType)that).resType));
        }

        private boolean sameArgTypes(FunType funType) {
            if (argTypes.length != funType.argTypes.length)
                return false;
            for (int i = 0; (i < argTypes.length); i++)
                if (!argTypes[i].isSameType(funType.argTypes[i]))
                    return false;
            return true;
        }

        boolean isSubType(Type that) {
            return (that == BAD_TYPE) ||
                   (that == ANY_TYPE) ||
                   ((that instanceof FunType) &&
                    ((FunType)that).subArgTypes(this) &&
                    resType.isSubType(((FunType)that).resType));
        }

        private boolean subArgTypes(FunType funType) {
            if (argTypes.length != funType.argTypes.length)
                return false;
            for (int i = 0; (i < argTypes.length); i++)
                if (!argTypes[i].isSubType(funType.argTypes[i]))
                    return false;
            return true;
        }

        public String toString() {
            StringBuffer buffer = new StringBuffer();
            buffer.append("(");
            for (int i = 0; i < argTypes.length; ++i) {
                if (i > 0)
                    buffer.append(", ");
                buffer.append(argTypes[i]);
            }
            buffer.append(") => ");
            buffer.append(resType);
            return buffer.toString();
        }
    }
}
